/**
 * mobios.app.BuildVPIndex.java 2006.06.25
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.06.25: Modified from jdb 1.0 VPBulkLoader.java, by Rui Mao
 */

package mobios.app;

import java.util.logging.Level;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;

import mobios.db.Table;
import mobios.db.TableManager;
import mobios.dist.Metric;
import mobios.dist.CountedMetric;
import mobios.index.algorithms.PartitionMethods;
import mobios.index.algorithms.PivotSelectionMethods;
import mobios.index.algorithms.PartitionMethod;
import mobios.index.algorithms.PivotSelectionMethod;
import mobios.index.VPIndex; // for javadoc


/**
 * This is the main entry point for building a VP Index. It can build either a single {@link VPIndex}, 
 * or a series of {@link VPIndex}es.
 * 
 * 
 * @author Rui Mao, Willard
 * @version 2006.07.26
 */
public class BuildVPIndex {

    String dataType = null;
    File mobiosParentDir = null;
    File sourceData = null;
    String indexName = null;
    PivotSelectionMethod psm = PivotSelectionMethods.valueOf("FFT");
    int pNum = 3;

    PartitionMethod dpm = PartitionMethods.valueOf("BALANCED");
    int sf = 3;

    int mls = 100;
    int pathLength = 0;
    Level debug = Level.OFF;

    int frag = -1;
    int dim = -1;
    boolean bucket = false;
    double maxR = 0.1;

    int initialSize = Integer.MAX_VALUE;
    int finalSize = Integer.MAX_VALUE;
    int stepSize = Integer.MAX_VALUE;
    
	protected BuildVPIndex(String[] args)
    {
        // final String indexExtName = ".index";

        for (int i = 0; i < args.length; i = i + 2) {
            if (args[i].equalsIgnoreCase("-t"))
                dataType = args[i + 1];
            
            else if (args[i].equalsIgnoreCase("-d")) 
                mobiosParentDir = new File(args[i+1]);
            
            else if (args[i].equalsIgnoreCase("-i"))
                sourceData = new File(args[i + 1]);
            
            else if (args[i].equalsIgnoreCase("-o"))
                indexName = args[i + 1];

            else if (args[i].equalsIgnoreCase("-psm"))
                psm = PivotSelectionMethods.valueOf(args[i + 1].toUpperCase());

            else if (args[i].equalsIgnoreCase("-p"))
                pNum = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-dpm"))
                dpm = PartitionMethods.valueOf(args[i + 1].toUpperCase());

            else if (args[i].equalsIgnoreCase("-f"))
                sf = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-pl"))
                pathLength = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-s")) {
                int size = Integer.parseInt(args[i + 1]);
                initialSize = size;
                finalSize = size;
                stepSize = size;

            } else if (args[i].equalsIgnoreCase("-m"))
                mls = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-sm"))
                initialSize = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-la"))
                finalSize = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-st"))
                stepSize = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-b"))
                bucket = (Integer.parseInt(args[i + 1]) == 1);

            else if (args[i].equalsIgnoreCase("-frag"))
                frag = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-dim"))
                dim = Integer.parseInt(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-r"))
                maxR = Double.parseDouble(args[i + 1]);

            else if (args[i].equalsIgnoreCase("-g"))
                debug = Level.parse(args[i + 1]);

            else
                throw new IllegalArgumentException("Invalid option " + args[i]);
        }

        dpm.setMaxRadius(maxR);
        dpm.setMaxLeafSize(mls);

    }

    /**
	 * The basic commandline options are:
	 * 
	 * -t [data type: one of "protein", "dna", "vector", "image", "ms", "msms"]
	 * -d [location of "mobiosData" directory; if directory does not exist, will be created at specified location]
	 * -i [input data file name]
	 * -o [output index file name]
	 * -psm [the pivot selection selection method: random, fft, center, pcaonfft, pca]
	 * -p [number of pivots in an index node]
	 * -dpm [data partition method: balanced, clusteringkmeans, clusteringboundary]
	 * -f [fanout of a pivot]
	 * -m [maximum number of children in leaf nodes]
	 * -pl [path length, default 0]
	 * -g [debug level]
	 * -frag [fragment length, only meaningful for sequences]
	 * -dim [dimension of vector data to load]
	 * -b [whether bucketing will be used, 1: use]
	 * -s [size of index]
	 * -r [maximum radius for partition]
	 * 
	 * For building multiple databases, use the following options:
	 * 
	 * -sm [size of smallest index]
	 * -la [size of largest index]
	 * -st [step size of index]
	 * 
	 * When using the advanced options, the size of the index will be appended to the given index name for each index.
	 * 
	 * The {@link Metric} is hardcoded for each data type.
	 * 
	 * @param args
	 */
	public static void main(String[] args) {
        BuildVPIndex bvpi = new BuildVPIndex(args);
        bvpi.batchBulkLoad();
    }


    private TableManager tableManager;
    
	/*
	 * See main() method for description of parameters
	 */
	protected void batchBulkLoad() {
		// statistics
		double startTime, endTime;
		final int indexNum = (finalSize - initialSize) / stepSize + 1;
		double[] buildTimes = new double[indexNum];
		int[] distCalNum = new int[indexNum];

        tableManager = TableManager.openTableManager(mobiosParentDir);
		
		for (int size = initialSize, i = 0; (size <= initialSize)
				& (i < indexNum); size += stepSize, i++) {
			Table table = null;
			try {
                if (dim != -1)
                    table = getTable(dataType, sourceData, size, dim);
                else
                    table = getTable(dataType, sourceData, size, frag);
			} catch (IOException e) {
				e.printStackTrace();
			}

			CountedMetric countMetric = new CountedMetric(table.getMetric());

			if (debug != Level.OFF)
				System.out.println(" building size:" + table.size() + "..., ");

			String indexIOName = indexName + "-" + table.size();

			if (!removeFilesWithHeader(indexIOName)) {
				System.out.println("Can not remove old files!");
				System.exit(-1);
			}

			startTime = System.currentTimeMillis();
			table.setMetric(countMetric);

			table.buildVPIndex(indexName, psm, pNum, dpm, sf, mls, pathLength,
					bucket, debug);

			endTime = System.currentTimeMillis();

			buildTimes[i] = (endTime - startTime) / 1000.00;
			distCalNum[i] = countMetric.getCounter();
			countMetric.clear();
		}

		// all indecies had been built, now out put the statistics.
		System.out
				.println("Database size, building time(seconds), #distance calculation:");
		for (int i = 0; i < indexNum; ++i)
			System.out.println((initialSize + stepSize * i) + ", "
					+ buildTimes[i] + ", " + distCalNum[i]);
	}

	protected Table getTable(String dataType, File fileName, int size, int dimOrFrag) throws IOException
    {
        return tableManager.createTable(dataType, fileName, size, dimOrFrag);
    }

    /*
	 * Remove all files with the given header, a utility method, not to be used
	 * by user.
	 * 
	 * @param header
	 *            header of file name, can have directory name, all files can be
	 *            expressed as header* are deleted
	 * @return true if runs sucessfully
	 */
	private boolean removeFilesWithHeader(String header) {
		if (header == null)
			return true;

		class HeaderFilter implements FileFilter {
			String header;

			HeaderFilter(File f) {
				header = f.getAbsolutePath();
			}

			public boolean accept(File pathName) {
				String input = pathName.getAbsolutePath();
				return input.startsWith(header);
			}
		}
		;

		File f = new File(header);
		File parent = (new File(f.getAbsolutePath())).getParentFile();
		// System.out.println("Parent:" + parent.getAbsolutePath() );

		File[] toDelete = parent.listFiles(new HeaderFilter(f));
		for (int i = 0; i < toDelete.length; i++) {
			if (!toDelete[i].delete()) {
				System.out.println("Could not delete file:"
						+ toDelete[i].getAbsolutePath());
				return false;
			}
			// System.out.println("File: " + toDelete[i].getAbsolutePath() + "
			// deleted.");
		}

		return true;
	}
}
