/**
 * mobios.db.DoubleVectorTable 2006.07.24 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 
 * 2006.07.24: Added, by Willard
 */
package mobios.db;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.logging.Logger;

import mobios.dist.LMetric;
import mobios.dist.Metric;
import mobios.type.DoubleVector;
import mobios.util.Debug;

/**
 * 
 * @author willard
 */
public class DoubleVectorTable extends Table {
	
    /**
     * 
     */
    private static final long   serialVersionUID = 7630078213101669086L;

    static final Metric DEFAULT_METRIC   = LMetric.EuclideanDistanceMetric;

    private int dimension;
    
    /**
     * The first line of the file should have two integers, separated by white space. The first is
     * the dimension of the DoubleVector, the second is the total number of data points. Each
     * following line is a DoubleVector, with each dimension separated by white space. The {@link Metric}
     * used is the DEFAULT_METRIC.
     * 
     * @param fileName
     *        the filename of the source file
     * @param maxSize
     *        max number of data points to read
     * @param dimNum
     *        number of dimensions to read from data.
     * @throws IOException
     */
    public DoubleVectorTable(File sourceData, int maxSize, int maxDimension) throws IOException {
        this(sourceData, maxSize, maxDimension, DEFAULT_METRIC);
    }

    /**
     * 
     * The first line of the file should have two integers, separated by white space. The first is
     * the dimension of the DoubleVector, the second is the total number of data points. Each
     * following line is a DoubleVector, with each dimension separated by white space.
     * 
     * @param sourceData
     *        the filename of the source data
     * @param maxSize
     *        max number of data points to read
     * @param dimNum
     *        number of dimensions to read from data.
     * @throws IOException
     */
    DoubleVectorTable(File sourceData, int maxSize, int maxDimension, Metric metric)
            throws IOException {
        super(sourceData, maxSize, metric);
        BufferedReader reader = new BufferedReader(new FileReader(sourceData));
        loadData(reader, maxSize, maxDimension);
    }

    /**
     * @param reader
     * @param maxSize
     * @param dimNum
     * @throws IOException
     */
    void loadData(BufferedReader reader, int maxSize, int maxDimension) throws IOException {

        String line;
        ArrayList<DoubleVector> doubleVectors = new ArrayList<DoubleVector>();
        ArrayList<Integer> originalRowIDsArrayList = new ArrayList<Integer>();
        // read vector values from file
        line = reader.readLine(); // read the first line
        if (line != null)
            line = line.trim();

        String[] metaData = line.split("[ \t]+");

        if (metaData.length != 2) {
            System.out.println("Error: Cannot parse the data file.");
            System.exit(-1);
        }

        final int dim = java.lang.Integer.parseInt(metaData[0]); // dimension
        if (dim < maxDimension || maxDimension <=0)
        		dimension = dim;
        else
        		dimension = maxDimension;
        	// size = java.lang.Integer.parseInt(metaData[1]); // total number of data

        if (Debug.debug) {
            Logger.getLogger("mobios.index").info("dim: " + dim);
            Logger.getLogger("mobios.index").info("size: " + maxSize);
        }

        int numData = 0;
        double[] data = new double[dimension];

        line = reader.readLine();
        if (line != null)
            line = line.trim();

        while (line != null && numData < maxSize) {
            String[] row = line.split("[ \t]+");

            if (Debug.debug) {
                for (int i = 0; i < row.length; i++)
                    Logger.getLogger("mobios.index").finer("row[" + i + "]: " + row[i]);
            }

            for (int i = 0; i < dimension; i++)
                data[i] = java.lang.Double.parseDouble(row[i]);

            // System.out.println(new DoubleVector(new Integer(numData), data));

            originalRowIDsArrayList.add(numData, numData);
            doubleVectors.add(new DoubleVector(this, numData, data));

            line = reader.readLine();
            if (line != null)
                line = line.trim();

            numData++;
        }
        doubleVectors.trimToSize();
        this.data = doubleVectors;

        originalRowIDs = new int[originalRowIDsArrayList.size()];
        for (int i = 0, e = originalRowIDsArrayList.size(); i < e; i++) {
            originalRowIDs[i] = originalRowIDsArrayList.get(i);
        }
    }
    
    public int getDimension() {
    	return dimension;
    }
    
}
