/**
 * mobios.db.ImageTable 2006.07.24 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 
 * 2006.07.24: Added, by Willard
 */
package mobios.db;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;

import mobios.dist.ImageMetric;
import mobios.dist.Metric;
import mobios.type.Image;


/**
 * @author Willard, Rui Mao
 * @version 2006.08.08
 */
public class ImageTable extends Table {
    private static final long   serialVersionUID = 1574357904350312666L;

    private static final Metric DEFAULT_METRIC   = new mobios.dist.ImageMetric();

    public ImageTable(File sourceDir, int maxSize) throws FileNotFoundException {
        this(sourceDir, maxSize, DEFAULT_METRIC);
    }

    /**
     * Load image data from file with max distance information. This data file contains floating
     * numbers for features of the images. Each line contains: Image file name, Feature value 1,
     * Feature value 2, ..., Feature value m.
     * 
     * @param reader
     *        the {@link BufferedReader} for the image data file.
     * @param readerMaxInfo
     *        the {@link BufferedReader} for the image max distance information.
     * @param maxSize
     *        represents max number of rows of data to be loaded.
     * @throws FileNotFoundException
     */
    ImageTable(File sourceDir, int maxSize, Metric metric) throws FileNotFoundException {
        super(sourceDir, maxSize, metric);

        BufferedReader reader, readerMaxInfo = null;
        reader = new BufferedReader(new FileReader(sourceDir + File.pathSeparator + "allfeas.dat"));
        try {
            readerMaxInfo = new BufferedReader(new FileReader(sourceDir + File.pathSeparator + "maxinfo.dat"));
        } catch (FileNotFoundException e1) {
            System.err.println("maxinfo.dat not found. loading data without Max information");
            loadData(reader, maxSize);
        }
        loadData(reader, readerMaxInfo, maxSize);
    }

    private void loadData(BufferedReader reader, int maxSize) {
        String line;
        int count = 0;
        ArrayList<Image> images = new ArrayList<Image>(imageNum);
        try {
            // read values from the text file.
            line = reader.readLine();
            if (line != null)
                line = line.trim();

            //TODO added by mory
            originalRowIDs = new int[maxSize];
            
            while (line != null && count < maxSize) {
                float[] aList = new float[totalFeaLen];
                String[] lineSegments = line.split(" ");
                for (int i = 1; i <= totalFeaLen; i++) {
                    aList[i - 1] = Float.parseFloat(lineSegments[i]);
                }
                images.add(new Image(this, count, aList));
                originalRowIDs[count] = count;
                count++;
                line = reader.readLine();
                if (line != null)
                    line = line.trim();
            }
        } catch (java.io.IOException e) {
            e.printStackTrace();
            throw new java.lang.IllegalStateException(
                    "Error occured when reading Image features file: " + reader);
        }
        images.trimToSize();
        data = images;

        //TODO added by mory
        if (images.size() <maxSize)
        {
            int [] temp = new int[images.size()];
            System.arraycopy(originalRowIDs, 0, temp, 0, images.size());
            originalRowIDs = temp;
        }
    }

    private void loadData(BufferedReader reader, BufferedReader readerMaxInfo, int maxSize) {
        String line;
        String lineMaxInfo;
        int count = 0;
        ArrayList<Image> images = new ArrayList<Image>(imageNum);
        try {
            // read values from the text file.
            line = reader.readLine();
            if (line != null)
                line = line.trim();
            lineMaxInfo = readerMaxInfo.readLine();
            if (lineMaxInfo != null)
                lineMaxInfo = lineMaxInfo.trim();

            //TODO added by mory
            originalRowIDs = new int[maxSize];
            
            while (line != null && lineMaxInfo != null && count < maxSize) {
                float[] aList = new float[totalFeaLen];
                double[] maxDist = new double[feaLength.length];
                String[] lineSegments = line.split(" ");
                String[] lineMaxInfoSegments = lineMaxInfo.split(" ");
                for (int i = 1; i <= totalFeaLen; i++) {
                    aList[i - 1] = Float.parseFloat(lineSegments[i]);
                }
                for (int i = 0; i < feaLength.length; i++) {
                    maxDist[i] = java.lang.Double.parseDouble(lineMaxInfoSegments[i]);
                }
                images.add(new Image(this, count, aList, maxDist));
                originalRowIDs[count] = count;
                count++;
                line = reader.readLine();
                if (line != null)
                    line = line.trim();
                lineMaxInfo = readerMaxInfo.readLine();
                if (lineMaxInfo != null)
                    lineMaxInfo = lineMaxInfo.trim();
            }
        } catch (java.io.IOException e) {
            e.printStackTrace();
            throw new java.lang.IllegalStateException(
                    "Error occured when reading Image features file: " + reader);
        }
        images.trimToSize();
        
        //TODO added by mory
        if (images.size() <maxSize)
        {
            int [] temp = new int[images.size()];
            System.arraycopy(originalRowIDs, 0, temp, 0, images.size());
            originalRowIDs = temp;
        }
        
        data = images;
    }

    /**
     * A utility method to create max distance information for each image. Output has a row for each
     * image, which corresponds to the maximum distance of this image to the other images.
     */
    public void createMaxInfo(List r) {
        ImageMetric metric = new ImageMetric();
        int sz = r.size();
        double maxDist;
        double tempVal;
        for (int k1 = 0; k1 < sz; k1++) {
            Image ob1 = (Image) r.get(k1);
            for (int FeaIndex = 0; FeaIndex < feaLength.length; FeaIndex++) {
                maxDist = 0.0;
                for (int k2 = 0; k2 < sz; k2++) {
                    Image ob2 = (Image) r.get(k2);
                    tempVal = metric.getDistance_Fea(ob1, ob2, FeaIndex);
                    if (tempVal > maxDist) {
                        maxDist = tempVal;
                    }
                }
                System.out.print(maxDist);
                System.out.print(" ");
            }
            System.out.println();
        }
    }

    final public int totalFeaLen = 66;

    final public int feaLength[] = { 3, 48, 15 };

    final public int imageNum    = 10221;
}
