/**
 * mobios.db.Table 2006.07.24 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 
 * 2006.07.24: Added, by Willard
 */
package mobios.db;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

import mobios.dist.Metric;
import mobios.index.AbstractIndex;
import mobios.index.Index;
import mobios.index.VPIndex;
import mobios.index.algorithms.PartitionMethod;
import mobios.index.algorithms.PivotSelectionMethod;
import mobios.type.IndexObject;

/**
 * Represents a set of data points in a database, including it's corresponding metric and any
 * indexes that have been built over it.
 * 
 * @author Willard
 */
public abstract class Table implements Serializable {

    protected File                      	sourceData;
    protected int                         maxSize;
    protected String                      indexFileName;
    protected Metric                      metric;
    protected AbstractIndex[]             index;
    int[]                                 originalRowIDs;
    protected transient List<? extends IndexObject> data;
    private int                           tableLocation = -1;
    

    /**
     * Constructor.
     * 
     * @param sourceData
     *        the name of the source data file.
     * @param maxDataSize
     *        the maximum number of data points to read from the file.
     * @param metric
     *        the metric to use when comparing two datapoints or building an index.
     */
    protected Table(File sourceData, int maxSize, Metric metric) {
        if (sourceData == null)
            throw new IllegalArgumentException("sourceData cannot be null!");
        if (maxSize <= 0)
            throw new IllegalArgumentException("maxDataSize must be greater than zero!");
        if (metric == null)
            throw new IllegalArgumentException("Metric cannot be null!");
        this.sourceData = new File(sourceData.getName());
        this.maxSize = maxSize;
        this.metric = metric;
        this.index = new AbstractIndex[1];
    }

    /**
     * @return the number of datapoints in the table.
     */
    public int size() {
        if (index[0] == null)
            return data.size();
        else
            return index[0].size();
    }

    /**
     * @return the name of the file used to construct this table.
     */
    public File getSourceDataFile() {
        return sourceData;
    }

    /**
     * @return the list of datapoints in this table.
     */
    public List<? extends IndexObject> getData() {
        if (index[0] == null)
            return data;
        else 
            return index[0].getAllPoints();
    }

    /**TODO
     * @param rowID
     * @return
     */
    public int getOriginalRowID(int rowID) {
        return originalRowIDs[rowID];
    }

    /**TODO
     * @return the index built over this table.
     */
    public Index getIndex() {
        return index[0];
    }

    /**TODO
     * @return the metric used when comparing two objects in this table.
     */
    public Metric getMetric() {
        return metric;
    }

    /**TODO
     * @param metric
     */
    public void setMetric(Metric metric) {
        this.metric = metric;
    }

    // TODO improve this code
    /**
     * TODO
     */
    public void compressData() {
        // first sort the list according to the data points.
        Collections.sort(data);

        // then, make a list of the unique dataPoints.
        final int dataSize = data.size();
        ArrayList<IndexObject> compressedData = new ArrayList<IndexObject>(dataSize);
        int[] rowIDs2 = new int[dataSize];

        IndexObject dataPoint1 = data.get(0);
        int tempSize = 1;

        IndexObject dataPoint2;
        for (int i = 1; i < dataSize; i++) {
            dataPoint2 = (IndexObject) data.get(i);
            if (dataPoint1.equals(dataPoint2)) {
                tempSize++;
            } else {
                if (tempSize > 1) {
                    for (int j = i - tempSize; j < i; j++) {
                        int rowID = data.get(j).getRowID();
                        rowIDs2[j] = originalRowIDs[rowID];
                    }
                    dataPoint1.setRowID(i - tempSize);
                    dataPoint1.setRowIDLength(tempSize);
                } else {
                    int rowID = data.get(i - 1).getRowID();
                    rowIDs2[i - 1] = originalRowIDs[rowID];
                    dataPoint1.setRowID(i - 1);
                }
                compressedData.add(dataPoint1);
                dataPoint1 = dataPoint2;
                tempSize = 1;
            }
        }

        if (tempSize > 1) {
            for (int i = dataSize - tempSize; i < dataSize; i++) {
                int rowID = data.get(i).getRowID();
                rowIDs2[i] = originalRowIDs[rowID];
            }
            dataPoint1.setRowID(dataSize - tempSize);
            dataPoint1.setRowIDLength(tempSize);
        } else {
            int rowID = data.get(dataSize - 1).getRowID();
            rowIDs2[dataSize - 1] = originalRowIDs[rowID];
            dataPoint1.setRowID(dataSize - 1);
        }
        compressedData.add(dataPoint1);

        compressedData.trimToSize();
        System.out.println("original size: " + dataSize + " compressed data size: "
                + compressedData.size());
        data = compressedData;
        originalRowIDs = rowIDs2;
    }

    /**TODO
     * @param psm
     * @param numPivots
     * @param pm
     * @param singlePivotFanout
     * @param maxLeafSize
     * @param maxPathLength
     * @param bucket
     * @param debugLevel
     */
    public void buildVPIndex(String indexFileName, PivotSelectionMethod psm, int numPivots, PartitionMethod pm,
            int singlePivotFanout, int maxLeafSize, int maxPathLength, boolean bucket,
            Level debugLevel) {

    	//check to see if index already exists
        TableManager tableManager = TableManager.getTableManager();
        if (tableManager.indexExists(indexFileName)) {
        		tableManager.close();
        		return;
        }
        tableLocation = tableManager.getLocation();
        File mobiosDir = tableManager.close();
        
        double startTime, endTime, buildTime;

        startTime = System.currentTimeMillis();
        this.indexFileName = indexFileName;
        
        // step 1: compress the data list
        if (bucket == true) {
            compressData();
        }
        index[0] = new VPIndex(mobiosDir, indexFileName, data, metric, psm, numPivots, pm, singlePivotFanout,
                maxLeafSize, maxPathLength, debugLevel);
        System.out.println("Created index...\n");

        endTime = System.currentTimeMillis();
        buildTime = (endTime - startTime) / 1000.00;
        System.out.println("Time to create index = " + buildTime + "...\n");

        tableManager = TableManager.openTableManager(mobiosDir.getParentFile());
        tableManager.putTable(this, tableLocation);
        tableManager.close();
    }

    /**TODO
     * @return
     */
    public int getTableLocation() {
        return tableLocation;
    }

    /**TODO
     * @return
     */
    public int getMaxSize() {
        return maxSize;
    }

    /**TODO
     * @return
     */
    public String getIndexFileName() {
        return indexFileName;
    }

	public int[] getOriginalRowIDsArray() {
		return originalRowIDs;
	}
	
	public void initTable(File mobiosDir) {
		for (int i = 0; i < index.length; i++) {
			index[i].initAndOpenOIOM(mobiosDir, true);
		}
	}
}
