/**
 * mobios.db.TableManager 2006.07.24 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 
 * 2006.07.24: Added, by Willard
 */
package mobios.db;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Hashtable;

import mobios.util.MckoiObjectIOManager;
import mobios.util.ObjectIOManager;

/**
 * The TableManager manages a set of {@link Table}s. The code is similar in structure to a DBMS. The
 * TableManager is used to create {@link Table}s. Then an {@link Index} is created over the Table.
 * 
 * @author Willard
 *
 */
public class TableManager implements Serializable {
	/**
	 * 
	 */
	private static final long serialVersionUID = -6911571514920379964L;

	private File mobiosDBDir;
	
	private static TableManager tableManager;

	private int count;

	private Hashtable<String, Integer> tableAddresses;

	private Hashtable<Integer, Long> oiomLocations;

	transient private Hashtable<Integer, Table> tables;

	transient protected ObjectIOManager oiom;

	private TableManager() {
		tableAddresses = new Hashtable<String, Integer>();
		oiomLocations = new Hashtable<Integer, Long>();
		tables = new Hashtable<Integer, Table>();
	}

	public static TableManager getTableManager() {
		if (tableManager == null)
			throw new Error("table manager has not been opened!");
		return tableManager;
	}

	/**
	 * 
	 * @param mobiosDir
	 * @return
	 */
	public static TableManager openTableManager(File mobiosParentDir) {
		if (tableManager == null) {
			File mobiosDBDir = new File(mobiosParentDir, "mobiosData");
			if (mobiosDBDir.exists()) {
				ObjectInputStream objectStream;
				try {
					objectStream = new ObjectInputStream(new FileInputStream(mobiosDBDir+File.separator+"mobios.db"));
					tableManager = (TableManager) objectStream.readObject();
				} catch (IOException e) {
					e.printStackTrace();
				} catch (ClassNotFoundException e) {
					e.printStackTrace();
				}
			} else {
				//create dir
				mobiosDBDir.mkdir();
				tableManager = new TableManager();
			}
			tableManager.mobiosDBDir = mobiosDBDir;
			tableManager.initAndOpenOIOM(false);
		}
		return tableManager;
	}

	/**
	 * TODO
	 * 
	 * @param readOnly
	 */
	protected void initAndOpenOIOM(boolean readOnly) {
		oiom = new MckoiObjectIOManager((new File(mobiosDBDir,"mobios")).getAbsolutePath(), "000", 1024 * 1024 * 1024,
				"Java IO", 4, 128 * 1024, readOnly);
		try {
			if (!oiom.open()) {
				throw new Error("Cannot open store for TableManager mobios.000");
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Test to see if the given index exists.
	 * 
	 * @param indexName the name of an index
	 * @return true, if the index exists; otherwise, false.
	 */
	public boolean indexExists(String indexName) {
		Integer tableLocation = tableAddresses.get(indexName);
		if (tableLocation != null) {
			System.out.println("The index exists.");
			return true;
		}
		return false;
	}
	
	/**
	 * Return a Table that an Index is defined over, given the indexName.
	 * 
	 * @param indexName the name of an {@link Index}
	 * @return the {@link Table} that the {@link Index} is defined over.
	 */
	public Table getTable(String indexName) {
		Table table = null;
		// if the table exists
		if (indexExists(indexName)) {
			double startTime, endTime, runTime;
			System.out.println("Deserializing table...");
			startTime = System.currentTimeMillis();
			try {
				table = (Table) oiom.readObject(oiomLocations.get(tableAddresses.get(indexName)));
				table.initTable(mobiosDBDir);
				tables.put(table.getTableLocation(), table);
			} catch (IOException e) {
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}

			endTime = System.currentTimeMillis();
			runTime = (endTime - startTime) / 1000.00;
			System.out.println("Time to deserialize table: " + runTime);
		}
		return table;
	}

	/**
	 * Returns a table given its index in the table manager.
	 * 
	 * @param tableIndex the index of the {@link Table} in the {@link TableManager}.
	 * @return the table, if it exists. otherwise, null.
	 */
	public Table getTable(int tableIndex) {
		if (!tables.containsKey(tableIndex)) {
			try {
				tables.put(tableIndex, (Table) oiom.readObject(oiomLocations
						.get(tableIndex)));
			} catch (IOException e) {
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				e.printStackTrace();
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
		}
		return tables.get(tableIndex);
	}

	/**
	 * Reserves an integer location in the table for a table. 
	 * 
	 * @return
	 */
	public synchronized int getLocation() {
		count++;
		return count;
	}

	/**
	 * 
	 * @return the size of the table manager.
	 */
	public long size() {
		try {
			return oiom.size();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return -1;
	}

	/**
	 * Put the given table at the given location in the TableManager.
	 * 
	 * @param table the {@link Table} to add into the TableManager.
	 * @param tableLocation the integer location the table belongs at. This number is obtained using getLocation().
	 * @return
	 */
	public long putTable(Table table, int tableLocation) {
		Long pointer = null;
		try {
			pointer = oiom.writeObject(table);
		} catch (IOException e) {
			e.printStackTrace();
		}
		// now add this address
		tableAddresses.put(table.getIndexFileName(), tableLocation);
		oiomLocations.put(tableLocation, pointer);

		return pointer;
	}

	protected void finalize() throws Throwable {
		try {
			close(); // close open files
		} finally {
			super.finalize();
		}
	}

	public File close() {
		File mobiosDir = null;
		try {
			ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(mobiosDBDir+File.separator+"mobios.db"));
			out.writeObject(tableManager);
			out.flush();
			out.close();
			oiom.close();
			mobiosDir = tableManager.mobiosDBDir;
			tableManager = null;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return mobiosDir;
		
	}


	/**
	 * TODO
	 * 
	 * @param dataType
	 * @param queryData
	 * @param maxDataSize
	 * @return
	 * @throws IOException
	 */public Table createTable(String dataType, File queryData, int maxDataSize) throws IOException {
		return createTable(dataType, queryData, maxDataSize, -1);
	}
	
	/**
	 * TODO
	 * 
	 * @param dataType
	 * @param fileName
	 * @param maxDataSize
	 * @param fragLength
	 * @param dim
	 * @return
	 * @throws IOException
	 */
	public Table createTable(String dataType, File fileName, int maxDataSize, int fragLengthOrDim) throws IOException {
		Table table = null;

		if (dataType.equalsIgnoreCase("ms")) {
			table = new SpectraTable(fileName, maxDataSize);
		} else if (dataType.equalsIgnoreCase("msms")) {
			table = new TandemSpectraTable(fileName, maxDataSize);
		} else if (dataType.equalsIgnoreCase("dna")) {
			table = new DNATable(fileName, maxDataSize, fragLengthOrDim);
		} else if (dataType.equalsIgnoreCase("rna")) {
			table = new RNATable(fileName, maxDataSize, fragLengthOrDim);
		} else if (dataType.equalsIgnoreCase("peptide")) {
			table = new PeptideTable(fileName, maxDataSize, fragLengthOrDim);
		} else if (dataType.equalsIgnoreCase("vector")) {
			table = new DoubleVectorTable(fileName, maxDataSize, fragLengthOrDim);
		} else if (dataType.equalsIgnoreCase("image")) {
			table = new ImageTable(fileName, maxDataSize);
        } else if (dataType.equalsIgnoreCase("integer")) {
		} else {
			throw new Error("Invalid data type!");
		}
		
		//put table in the table manager
		return table;
	}
	
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}

	private void readObject(java.io.ObjectInputStream in) throws IOException,
			ClassNotFoundException {
		in.defaultReadObject();
		tables = new Hashtable<Integer, Table>();
	}

	public Object clone() throws CloneNotSupportedException {
		throw new CloneNotSupportedException();
	}
	
}