/**
 * mobios.dist.ImageMetric 2006.06.19
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Change Log:
 * 2006.06.19: Copied from jdb 1.0, by Rui Mao
 */

package mobios.dist;

import mobios.type.Image;
import mobios.type.IndexObject;

/**
 * Computes the distance between images. Designed for the MoBIoS image dataset,
 * <code>ImageMetric</code> may not work for other datasets.
 * 
 * @author Wenguo Liu, Rui Mao, Willard
 * @version 2005.10.31
 */

public class ImageMetric implements Metric {
    /**
     * 
     */
    private static final long serialVersionUID = 3442536718895037166L;

    /**
     * Feature number for each image. We have three features for each image: structure, texture, and
     * histogram.
     */
    final private int         feaNum           = 3;

    /**
     * Number of floats that can represent each feature.
     */
    final private int         feaLength[]      = { 3, 48, 15 };

    /**
     * Distance function selection for each feature.
     */
    final private boolean     minBool[]        = { false, false, false };

    /**
     * The weights of each feature in the computation of distance.
     */
    final private double      weights[]        = { 0.333333, 0.333333, 0.333333 };

    /**
     * The max distance for each feature.
     */
    final private double      maxDist[]        = { 1.0, 60.0, 1.0 };

    /*
     * (non-Javadoc)
     * 
     * @see mobios.dist.Metric#getDistance(mobios.type.IndexObject, mobios.type.IndexObject)
     */
    public double getDistance(IndexObject one, IndexObject two) {
        return getDistance((Image) one, (Image) two);
    }

    /**
     * Computes the distance between two {@link Image}s
     * 
     * @param one
     *        the first {@link Image} to compute the distance on.
     * @param two
     *        the second {@link Image} to compute the distance on.
     * @return the distance between the two {@link Image}s
     */
    public double getDistance(Image one, Image two) {
        double dist = 0.0;
        for (int i = 0; i < feaNum; i++) {
            dist += (getDistance_Fea(one, two, i) / maxDist[i] * weights[i]);
            // dist += ( 2.0*getDistance_Fea(one, two, i) / (one.getMaxDist(i)+two.getMaxDist(i)) *
            // weights[i] ) ;
        }
        return dist;
    }

    /**
     * Computes the distance between two images using the given feature index.
     * 
     * @param one
     *        the first {@link Image} to compute the distance on.
     * @param two
     *        the second {@link Image} to compute the distance on.
     * @param FeaIndex
     *        the feature on which distance is to be computed.
     * @return the distance between the two {@link Image}s
     */
    public double getDistance_Fea(Image one, Image two, int FeaIndex) {
        int StartIndex = 0, EndIndex = 0, cnt;
        double dist = 0.0, tempval = 0.0;

        for (int i = 0; i < FeaIndex; i++)
            StartIndex += feaLength[i];
        EndIndex = StartIndex + feaLength[FeaIndex] - 1;

        // The first method for computing image object distance.
        if (minBool[FeaIndex]) {
            for (cnt = StartIndex; cnt <= EndIndex; cnt++) {
                dist += Math.min(one.getFeature(cnt), two.getFeature(cnt));
                tempval += one.getFeature(cnt);
                // to make it a symmetric Metric space, add the following line
                // tempval += two.m_Feas[cnt] ;
            }
            dist = Math.abs(1.0 - (dist / tempval));
        } else { // The second method for computing image object distance.
            for (cnt = StartIndex; cnt <= EndIndex; cnt++) {
                tempval = (one.getFeature(cnt) - two.getFeature(cnt));
                dist += (tempval * tempval);
            }
            dist = Math.sqrt(dist);
        }

        return dist;
    }

}
