/**
 * mobios.dist.MSMSMetric 2004.03.17
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Change Log:
 * 2004.05.31; Modified for performance, replaced ArrayLists with arrays, by Willard
 * 2004.08.02; Modified to implement MSMSConstants, by Smriti Ramakrishnan
 */

package mobios.dist;

import mobios.dist.Metric;
import mobios.type.DoubleVector;
import mobios.type.IndexObject;
import mobios.type.TandemSpectra;
import static mobios.dist.MSMSConstants.*;

/**
 * MSMSMetric is an implementation of a fuzzy cosine distance metric for comparing tandem spectra
 * signatures. Within a given tolerance, elements of the {@link TandemSpectra} are equal.
 * 
 * @author Smriti Ramakrishnan, Willard
 * @version 2006.07.31
 */
public class MSMSMetric implements Metric {
    /**
     * 
     */
    private static final long    serialVersionUID = 8368326281379099335L;
    
    /**
     * 
     * @param min
     * @param max
     * @param step
     * @param tol
     */
    public MSMSMetric(int min, int max, double step, double tol) {
        this.min = min;
        this.max = max;
        this.step = step;
        this.tol = tol;

        mscosdist = 0.0;
        absMassDiff = 0.0;
        massDiffTerm = 0.0;
    }

    /**
     * Default constructor provides default values for min, max, step, and tolerance:
     * 
     * min = 0; 
     * max = 0;
     * step = 0; 
     * tol = MSMSConstants.MS_TOLERANCE
     */
    public MSMSMetric() {
        this.min = 0;
        this.max = 0;
        this.step = 0;
        this.tol = MS_TOLERANCE;

        mscosdist = 0.0;
        absMassDiff = 0.0;
        massDiffTerm = 0.0;
    }

    /* (non-Javadoc)
     * @see mobios.dist.Metric#getDistance(mobios.type.IndexObject, mobios.type.IndexObject)
     */
    public double getDistance(IndexObject v1, IndexObject v2) {
        return getDistance((TandemSpectra) v1, (TandemSpectra) v2);
    }

    /**
     * Computes the distance between two {@link TandemSpectra}es.
     * 
     * @param v1
     *        the first {@link TandemSpectra} to compute distance on
     * @param v2
     *        the second {@link TandemSpectra} to compute distance on
     * @return the distance between the two {@link TandemSpectra}es
     */
    public double getDistance(TandemSpectra v1, TandemSpectra v2) {
        // now compute cosine distance
        double cos = getCosine(v1, v2);

        // from BinaryVectorMetric:getDistance()
        // if ( Math.abs(cos) > 1) {; Smriti

        if (Math.abs(Math.abs(cos) - 1) < COS_THRESHOLD) { // precision
            if (cos > 0)
                cos = 1; // very similar
            else
                cos = -1; // very unsimilar
        } else if (Math.abs(cos) > 1) {
            System.out.println("COS_THRESHOLD = " + COS_THRESHOLD + ", cos = " + cos);
            System.out.println("got cosine > 1, cosine=" + cos + ", :" + v1.toString() + ", v2:"
                    + v2.toString() + "Quitting.");
            System.exit(0);
        }
        // }

        /*
         * System.out.println("MSTOL = " + tol); System.out.println("MSTOL = " +
         * MS_PRECURSOR_TOLERANCE); System.out.println("MSDKO-1 : " + v1);
         * System.out.println("MSDKO-2 : " + v2);
         */

        // ********************************************
        mscosdist = Math.acos(cos);
        massDiffTerm = getAbsPrecursorMassDiff(v1, v2);
        double dist = massDiffTerm + mscosdist;
        // ********************************************

        /*
         * System.out.println("K1 = " + v1); System.out.println("K2 = " + v2);
         * System.out.println("mscosDist = " + mscosdist + ", Cosine = " + cos + ", MassDiff = " +
         * massDiff); System.out.println("Total Dist = " + dist); System.out.println();
         */
        return dist;

        // smriti - 30 sep
        // return (1 - cos);
        /*
         * //extended jaccard double[] one = v1.getData(); double[] two = v2.getData(); double spc =
         * getInnerProduct(one, two); double dist = 1 - (spc / (getMagnitude(one) +
         * getMagnitude(two) - spc)); return dist;
         */
        // getCosine(,v2));//weijia
    }

    /**
     * Computes the absolute difference between precursor masses - within a tolerance
     * @param v1
     * the first {@link TandemSpectra} to compute absolute difference on
     * @param v2
     * the second {@link TandemSpectra} to compute absolute difference on
     * @return the absolute difference between precursor masses of the respective {@link TandemSpectra}es
     */
    private double getAbsPrecursorMassDiff(TandemSpectra v1, TandemSpectra v2) {
        double m1 = v1.getPrecursorMass();
        double m2 = v2.getPrecursorMass();

        absMassDiff = Math.abs(m1 - m2);

        if (absMassDiff < COS_THRESHOLD)
            absMassDiff = 0.0;

        if (absMassDiff <= MS_PRECURSOR_TOLERANCE)
            return 0.0;
        else
            return (absMassDiff);
    }

    /**
     * @param one
     * @param two
     * @return
     */
    private double getCosine(DoubleVector one, DoubleVector two) {
        double[] v1 = one.getData();
        double[] v2 = two.getData();

        return getInnerProduct(v1, v2) / (getMagnitude(v1) * getMagnitude(v2));

        // smriti - 30sep
        /*
         * double numer = getInnerProduct(v1, v2) ; double denom = getMagnitude(v1) *
         * getMagnitude(v2); double denom = getMagnitude(v1) * getMagnitude(v1); double d2 =
         * getMagnitude(v2) * getMagnitude(v2); denom = (denom + d2) / 2; return numer/denom;
         */
    }

    /**
     * Computes inner product within a certain tolerance-- "fuzzy" inner product.
     */
    private int getInnerProduct(double[] v1, double[] v2) {
        int dist = 0;
        int i = 0, j = 0;
        double val1;
        double val2;

        // System.out.println("TOL = " + tol);
        while (i < v1.length && j < v2.length) {
            val1 = v1[i];
            val2 = v2[j];
            // new version -- gets rid of Math.abs()
            if (val1 <= val2) {
                if (val2 - val1 <= tol) {
                    dist++;
                    i++;
                    j++;
                } else {
                    i++;
                }
            } else {
                if (val1 - val2 <= tol) {
                    dist++;
                    i++;
                    j++;
                } else {
                    j++;
                }
            }

            /*
             * old code if (Math.abs(val1 - val2) <= tol) { dist++; i++; j++; } else { if (val1 <
             * val2) i++; else j++; }
             */
            // debug
            // System.out.println("i = " + i + ", j = " + j);
        }
        // System.out.println("INNER PROD = " + dist);
        return dist;
    }

    /**
     * For experiment purpose only Method to compute inner product, cosine value and distance of v2
     * {@link VectorData} objects by treating them as binary vectors in a multidimensional space.The
     * Value will be stored in a {@link double[]} with the correspoding order.
     * 
     * @param -
     *        {@link VectorData}
     * @param v2 -
     *        {@link vectorData}
     * @return {@link double []}
     */
    /*
     * public double[] getAll(DataVector one, DataVector two) { double[] v1 = one.getData();
     * double[] v2 = two.getData(); double[] result = new double[3]; result[0] = getInnerProduct(v1,
     * v2); result[1] = result[0] / (getMagnitude(v1) * getMagnitude(v2)); result[2] =
     * Math.(result[1]); return result; }
     */

    /**
     * @param {@link VectorData}
     */

    /**
     * Computes the magnitude of the double array by treating it as a *sparse* binary vector (store
     * only 1's)
     * 
     * @param d
     * @return
     */
    private double getMagnitude(double[] d) {
        return Math.sqrt(d.length);
        // return d.length;
    }

    /**
     * Convert information of the instance into a {@link String} that contains the number of
     * dimension and meaning for each dimension
     */
    public String printDistance(TandemSpectra k1, TandemSpectra k2) {
        // set values of mscosdist and massDiff
        getDistance(k1, k2);
        java.text.DecimalFormat frm = new java.text.DecimalFormat("####.########");

        StringBuffer outStr = new StringBuffer(20);
        // outStr.append("k1: " + k1);
        // outStr.append("k2: " + k2 + "\n");
        outStr
                .append("MSCOSDIST = " + frm.format(mscosdist) + ", MASS_DIFF_TERM = "
                        + frm.format(massDiffTerm) + " (abs mass diff = " + frm.format(absMassDiff)
                        + ")\n");

        return outStr.toString();
    }

    // legacy
    int                  min, max;                               // min, max are mass ranges
    double               step;                                   // Same as Tolerance ?
    // end-legacy

    double               mscosdist;
    double               absMassDiff;
    double               massDiffTerm;

    double               tol;
    private final double COS_THRESHOLD    = 0.00005;
}
