/**
 * mobios.index.AbstractIndex 2006.05.09
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.05.09: Created, by Rui Mao, Willard
 */

package mobios.index;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import mobios.dist.Metric;
import mobios.type.IndexObject;
import mobios.util.MckoiObjectIOManager;
import mobios.util.ObjectIOManager;

/**
 * The primary interface for a distance-based index. Through AbstractIndex, the user can build a database
 * index, or read a pre-built index from a file, and then do the search.
 * 
 * @author Rui Mao, Willard
 * @version 2006.07.24
 */
public abstract class AbstractIndex implements Index {
    static final long                   serialVersionUID       = 2659219513493819541L;

    // name of the index node file
    protected String          fileName;
    transient protected static String   nodeFileNameExtension  = "000";
    transient protected ObjectIOManager oiom;
    transient protected Logger          logger;
    // metric that this index is built over.
    protected Metric                    metric;
    // total number of data objects
    protected int                       total_size;
    // address of the index root node in the index node file
    protected long                      root;

    // private to disallow creation using the default constructor
    private AbstractIndex() {
    }

    /**
     * The constructor in common use. Initializes the directory to store index files, opens up the 
     * {@link ObjectIOManager}, and defines the metric and debug level to use when building an index.
     * 
     * @param fileName
     * @param data
     * @param metric
     * @param debugLevel
     */
    public AbstractIndex(File mobiosDataDir, String fileName, List<? extends IndexObject> data, Metric metric,
            Level debugLevel) {
        if (fileName == null)
            throw new IllegalArgumentException("fileName cannot be null!");
        if (metric == null)
            throw new IllegalArgumentException("metric cannot be null!");
        if (data == null)
            throw new IllegalArgumentException("data list cannot be null!");
        if (debugLevel == null) {
            System.err.println("using default Debug Level: java.util.logging.Level.INFO");
            debugLevel = Level.INFO;
        }

        this.fileName = fileName;
        this.metric = metric;
        total_size = data.size();

        // 1. create the io manager, writeable
        initAndOpenOIOM(mobiosDataDir, false);

        // 2. create the logger
        logger = Logger.getLogger("mobios.index");
        // set the logger level;
        logger.setLevel(debugLevel);
        // init the Handler: currently use a console handler, in the future will probably use the
        // FileHandler, or both
        logger.addHandler(new ConsoleHandler());
        /*
         * try { logger.addHandler(new FileHandler(fileName+".log")); } catch (SecurityException e) {
         * e.printStackTrace(); } catch (IOException e) { e.printStackTrace(); }
         */
    }

    /**
     * Initializes and opens the {@link ObjectIOManager} for this index.
     * 
     * @param readOnly
     */
    public void initAndOpenOIOM(File dir, boolean readOnly) {
        oiom = new MckoiObjectIOManager(new File(dir, fileName).getAbsolutePath(), nodeFileNameExtension, 1024 * 1024 * 1024,
                "Java IO", 4, 128 * 1024, readOnly);

        String nodeFileName = fileName + "." + nodeFileNameExtension;
        try {
            if (!oiom.open()) {
                throw new Error("Cannot open store for: " + nodeFileName);
            }
            System.out.println("OIOM.size = " + oiom.size() + "\n");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.index.Index#getMetric()
     */
    public Metric getMetric() {
        return this.metric;
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.index.Index#size()
     */
    public int size() {
        return total_size;
    }


    /* (non-Javadoc)
     * @see mobios.index.Index#preLoad(int)
     */
    public void preLoad(int level) {
        double startTime = 0;
        double endTime = 0;
        long memoryUsed = 0;
        System.out.println("Start the preloading database into memory");
        Runtime rt = Runtime.getRuntime();
        memoryUsed = rt.totalMemory() - rt.freeMemory();
        startTime = System.currentTimeMillis();
        // hashing internal nodes first
        LinkedList<PreLoadTask> nodeList = new LinkedList<PreLoadTask>();
        PreLoadTask task = new PreLoadTask(root, 0);
        nodeList.add(task);
        int numPreLoadedNodes = 0;
        while (!nodeList.isEmpty()) {
            task = nodeList.removeFirst();
            Object node = null;
            try {
                node = oiom.readPersistObject(task.nodeAddress);
                numPreLoadedNodes++;
                if (task.currentLevel < level) {
                    if (node instanceof InternalNode) {
                        for (int i = 0; i < ((InternalNode) node).childAddresses.length; i++)
                            nodeList.add(new PreLoadTask(((InternalNode) node).childAddresses[i],
                                    task.currentLevel + 1));
                    }

                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        endTime = (System.currentTimeMillis() - startTime) / 1000;
        rt = Runtime.getRuntime();
        memoryUsed = rt.totalMemory() - rt.freeMemory() - memoryUsed;
        System.out.println("finish the preloading database into memory " + numPreLoadedNodes
                + " nodes loaded using " + endTime + " seconds " + memoryUsed + " byte memory");
    }

    private class PreLoadTask {
        private long nodeAddress;
        private int  currentLevel;

        private PreLoadTask(long nodeAddress, int currentLevel) {
            this.nodeAddress = nodeAddress;
            this.currentLevel = currentLevel;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.index.Index#getAllPoints()
     */
    public List<IndexObject> getAllPoints() {
        return Cursor.getAllPoints(oiom, root);
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.index.Index#close()
     */
    public void close() throws IOException {
        oiom.close();
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.index.Index#destroy()
     */
    public void destroy() throws IOException {
        oiom.close();
        // delete the oiom
        boolean success = (new File(fileName + nodeFileNameExtension)).delete();
        if (!success) {
            throw new IOException("Deletion failed.");
        }
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        out.defaultWriteObject();
    }

    private void readObject(ObjectInputStream in) throws IOException,
            ClassNotFoundException {
        in.defaultReadObject();
    }
}
