/**
 * mobios.index.Cursor 2006.05.09
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.05.09: Copied from the original mobios package, by Rui Mao
 */

package mobios.index;

import mobios.type.IndexObject;
import mobios.type.DoubleIndexObjectPair;
import mobios.util.ObjectIOManager;
import mobios.dist.Metric;

import java.util.List;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * Implements an {@link Iterator}, iterating on the results set.
 * 
 * @author Rui Mao, Willard
 * @version 2006.05.31
 */
public abstract class Cursor implements Iterator {
    /**
     * The {@link ObjectIOManager} for the index.
     * 
     */
    ObjectIOManager                   oiom;
    
    /**
     * The {@link metric} that the index is defined over.
     * 
     */
    Metric                            metric;
    
    /**
     * The address of the root node in the {@link ObjectIOManager}
     * 
     */
    long                              rootAddress;

    /**
     * Keeps a running list of the results of this search.
     */
    LinkedList<DoubleIndexObjectPair> result;

    /**
     * Defines a cursor over the given index.
     * 
     * @param index
     */
    protected Cursor(Index index) {
        this.oiom = index.getOIOM();
        this.metric = index.getMetric();
        this.rootAddress = index.getRootAddress();

        result = new LinkedList<DoubleIndexObjectPair>();
    }

    /* (non-Javadoc)
     * @see java.util.Iterator#hasNext()
     */
    public boolean hasNext() {
        if (result.size() > 0)
            return true;
        else {
            continueSearch();
            return (result.size() > 0);
        }
    }

    /* (non-Javadoc)
     * @see java.util.Iterator#next()
     */
    public DoubleIndexObjectPair next() {
        if (result.size() > 0)
            return expandNext();
        else {
            continueSearch();
            if (result.size() > 0)
                return expandNext();
            else
                throw new java.util.NoSuchElementException("The search is already done!");
        }
    }

    /*
     * called by next() to expand the IndexObject returned by the search.
     */
    private DoubleIndexObjectPair expandNext() {
        DoubleIndexObjectPair dioPair = result.remove();
        IndexObject iObject = dioPair.getObject();
        if (iObject instanceof IndexObject) {
            double _double = dioPair.getDouble();
            IndexObject[] cioArray = ((IndexObject) dioPair.getObject()).expand();
            // System.out.println("length=" + cioArray.length);
            for (int i = 0; i < cioArray.length; i++) {
                result.addFirst(new DoubleIndexObjectPair(_double, cioArray[i]));
            }
            return result.remove();
        }
        return dioPair;
    }

    /* (non-Javadoc)
     * @see java.util.Iterator#remove()
     */
    public void remove() {
        throw new UnsupportedOperationException();
    }

    /**
     * TODO
     */
    abstract void continueSearch();

    /**
     * Return a list of all the data points in the subtree rooted at the address
     * 
     * @param oiom
     * @param address
     * @return each data point is an IndexObject, consist of a row ID and an index key
     */
    static List<IndexObject> getAllPoints(ObjectIOManager oiom, long address) {
        IndexNode node = null;
        try {
            node = (IndexNode) oiom.readObject(address);
        } catch (Exception e) {
            e.printStackTrace();
        }

        LinkedList<IndexObject> result = new LinkedList<IndexObject>();
        for (int pivot = 0; pivot < node.numPivots(); pivot++) {
            result.add(node.getPivot(pivot));
        }

        if (node instanceof InternalNode) {
            InternalNode iNode = (InternalNode) node;
            for (int child = 0; child < iNode.numChildren(); child++)
                result.addAll(getAllPoints(oiom, iNode.getChildAddress(child)));
        } else if (node instanceof LeafNode) {
            LeafNode lNode = (LeafNode) node;
            for (int child = 0; child < lNode.numChildren(); child++) {
                result.add(lNode.getChild(child));
            }
        } else {
            throw new UnsupportedOperationException("Node type: " + node.getClass()
                    + " is not supported!");
        }

        // decompress
        LinkedList<IndexObject> expandedList = new LinkedList<IndexObject>();
        Iterator<IndexObject> p = result.iterator();
        while (p.hasNext()) {
            IndexObject o = p.next();
            if (o instanceof IndexObject) {
                IndexObject[] expanded = ((IndexObject) o).expand();
                for (IndexObject temp : expanded)
                    expandedList.addFirst(temp);
            } else
                expandedList.addFirst(o);
        }
        return expandedList;
    }

    /**
     * TODO
     * @param task
     */
    abstract protected void visit(QueryTask task);

    /**
     * TODO
     * @author Rui Mao, Willard
     *
     */
    protected class QueryTask {
    }
}
