/**
 * mobios.index.VPIndex 2006.05.09 
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 
 * 2006.05.09: Created, by Rui Mao, Willard
 * 2006.07.09: added KNN search, by Weijia Xu
 */

package mobios.index;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;

import mobios.dist.Metric;
import mobios.index.algorithms.PartitionMethod;
import mobios.index.algorithms.PartitionResults;
import mobios.index.algorithms.PivotSelectionMethod;
import mobios.index.Query;
import mobios.index.RangeQuery;
import mobios.type.IndexObject;
import mobios.util.Debug;
import mobios.util.ObjectIOManager;

/**
 * A Vantage Point Tree (VPT) index.
 * 
 * @author Rui Mao, Willard, Weijia Xu
 * @version 2006.05.09
 */
public class VPIndex extends AbstractIndex {
    private static final long    serialVersionUID = 6928847050089693231L;

    private PivotSelectionMethod psm;
    private int                  numPivot;
    private PartitionMethod      pm;
    private int                  singlePivotFanout;
    private int                  maxLeafSize;
    private int                  maxPathLength;

    private transient ArrayList<LoadTask>  taskList;

    /**
     * Builds an index over the specified Table.
     * 
     * @param fileName
     *        the fileName to give the index.
     * @param data
     *        the {@link List} of {@link IndexObject}s to build the index over 
     * @param metric
     *             the {@link Metric} to use when building the index.
     * @param psm
     *        the {@link PivotSelectionMethod} to use when building the index.
     * @param numPivot
     *        the number of pivots to use.
     * @param pm
     *        the {@link PartitionMethod} to use when building the index.
     * @param singlePivotFanout
     *        the fanout for a single pivot. the total fanout is singlePivotFanout^numPivot
     * @param maxLeafSize
     *        the maximum number of data points in a leaf node.
     * @param maxPathLength
     *        the number of previous distance calculations to store in a given node.
     * @param debugLevel
     *        the debug level.
     */
    public VPIndex(File mobiosDir, String fileName, List<? extends IndexObject> data, Metric metric, PivotSelectionMethod psm, int numPivot,
            PartitionMethod pm, int singlePivotFanout, int maxLeafSize, int maxPathLength, Level debugLevel) {

        // 1.a. initialize the object io manager
        super(mobiosDir, fileName, data, metric, debugLevel);

        // 1.b. initialize fields
        this.psm = psm;
        this.numPivot = numPivot;
        this.pm = pm;
        this.singlePivotFanout = singlePivotFanout;
        this.maxLeafSize = maxLeafSize;
        this.maxPathLength = maxPathLength;

        // 1.c. initialize two lists
        taskList = new ArrayList<LoadTask>();

        // 2. do the actual loading,
        this.root = load(data);

        // 4. set oiom to be read only
        try {
            close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        initAndOpenOIOM(mobiosDir, true);
    }

    /**
     * 
     * @param data
     * @return
     */private long load(List<? extends IndexObject> data) {
        // step 1: initialize the stack, put the root task into it.
        LoadTask task = new LoadTask(data, "0", -1, 0, new ArrayList<IndexObject>(
                maxPathLength));

        logger.info("building " + task.description + ", size: " + task.size + ", heap size: "
                + Runtime.getRuntime().totalMemory() / 1024 + "K, used size: "
                + (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1024
                + "K");

        // at the root, you always want to create an internal node

        // step 2.a: select pivots
        int numPivots = (numPivot >= task.size) ? task.size : numPivot;

        int[] pivots = psm.selectPivots(metric, task.compressedData, numPivots);

        // step 2.b: rearrange data list, put pivots at end:

        task.groupPivotsAtEnd(pivots);
        taskList.add(task);
        createAnInternalNode(task, pm, singlePivotFanout, maxLeafSize, 0);

        // step 2: do the DFS bulkloading
        int taskToGet;
        while ((taskToGet = taskList.size() - 1) > 0) {
            task = taskList.get(taskToGet);
            //System.out.println(task);
            
            if (task.node == null)
            {
                // step 2.a: select pivots
                numPivots = (numPivot >= task.size) ? task.size : numPivot;

                logger.finer("VP Selection Method:" + psm.toString());
                logger.finer("selecting VP using " + psm.toString() + " selection.., data size= "
                        + task.size + ", vp number= " + numPivots);

                pivots = psm.selectPivots(metric, task.compressedData, numPivots);

                // step 2.b: rearrange data list, put pivots at end:
                task.groupPivotsAtEnd(pivots);

                // step 2.c: create either an internal node or a leaf node depending on the maximum
                // leaf size allowed
                if (task.size - task.numPivots > maxLeafSize)
                    createAnInternalNode(task, pm, singlePivotFanout, maxLeafSize, taskToGet);
                else
                    createAndWriteLeafNode(taskList.remove(taskToGet));
            } else {
                writeInternalNode(taskList.remove(taskToGet));
            }
        }

        // step 3: save the root node to the ObjectIOManager
        return writeRootNode(taskList.remove(0));
    }

    private void createAnInternalNode(LoadTask task, PartitionMethod pm, int singlePivotFanout,
            int maxLeafSize, int nodeIndex) {
        logger.finest("Create an internal node...");

        // get pivots
        IndexObject[] pivots = task.getPivots();
        for (int i = 0; i < pivots.length; i++) {
            if (task.pivotHistoryList.size() < maxPathLength)
                task.pivotHistoryList.add(pivots[i]);
            else
                break;

        }

        PartitionResults partitionResults = pm.partition(metric, pivots, task.compressedData, 0,
                (task.size - task.numPivots), singlePivotFanout);

        int childrenNumber = partitionResults.size();

        // step 3. create an internal node, and save it to nodeList
        task.node = partitionResults.getInternalNode();

        // step 4: add tasks for each child into the stack
        for (int i = childrenNumber - 1; i >= 0; i--) {
            task.node.setChildAddress(i, -1);
            LoadTask newTask;
            // TODO why this?
            if (pivots.length < maxPathLength)
                newTask = new LoadTask(partitionResults.getPartition(i),
                        task.description + "." + i, nodeIndex, i, new ArrayList<IndexObject>(
                                task.pivotHistoryList));
            else
                newTask = new LoadTask(partitionResults.getPartition(i),
                        task.description + "." + i, nodeIndex, i, task.pivotHistoryList);
            taskList.add(newTask);
        }
    }

    private void createAndWriteLeafNode(LoadTask task) {
        IndexObject[] children = task.getDataPoints();
        IndexObject[] pivots = task.getPivots();

        double[][] distance = new double[children.length][task.numPivots];
        for (int i = 0; i < children.length; i++) {
            final IndexObject currentChild = children[i];
            for (int j = 0; j < task.numPivots; j++)
                distance[i][j] = metric.getDistance(currentChild, pivots[j]);
        }

        if (Debug.debug) {
            logger.finest("create a leaf node, distances(VP*children):");
            for (int i = 0; i < task.numPivots; i++) {
                for (int j = 0; j < children.length; j++)
                    logger.finest(distance[i][j] + ", ");
                logger.finest("");
            }
        }

        final int pivotHistoryListSize = task.pivotHistoryList.size();
        double[][] pathDistance = new double[children.length][pivotHistoryListSize];
        for (int i = 0; i < children.length; i++)
            for (int j = 0; j < pivotHistoryListSize; j++)
                pathDistance[i][j] = metric.getDistance(children[i], task.pivotHistoryList.get(j));

        try {
            VPLeafNode node = new VPLeafNode(pivots, children, task.size, distance, pathDistance);

            // write this node to the index file
            final long pointer = oiom.writeObject(node);
            InternalNode parentNode = taskList.get(task.parentNodeIndex).node;
            parentNode.setChildAddress(task.myIndex, pointer);
        } catch (Exception e) {
            System.out.print("Exception in creating HeavyLeafNode: " + e.toString());
            e.printStackTrace();
        }
    }

    private long writeInternalNode(LoadTask task) {
        final InternalNode iNode = (InternalNode) task.node;
        // check whether each child address has already been set.
        final int childNumber = iNode.numChildren();
        for (int j = 0; j < childNumber; j++)
            if (iNode.getChildAddress(j) == -1)
                System.out.println("current node: " + iNode.toString() + ", child " + j
                        + " is null!");

        // write current node, and set its parent
        long pointer = -1;
        try {
            pointer = oiom.writeObject(iNode);
        } catch (Exception e) {
            System.out.println("error writing index node to ObjectIOManager!");
            e.printStackTrace();
        }
        ;
        LoadTask parentNodeTask = taskList.get(task.parentNodeIndex);
        InternalNode parent = (InternalNode) parentNodeTask.node;
        parent.setChildAddress(task.myIndex, pointer);

        return pointer;
    }

    private long writeRootNode(LoadTask task) {
        final InternalNode iNode = (InternalNode) task.node;
        // check whether each child address has already been set.
        final int childNumber = iNode.numChildren();
        for (int j = 0; j < childNumber; j++)
            if (iNode.getChildAddress(j) == -1)
                System.out.println("current node: " + iNode.toString() + ", child " + j
                        + " is null!");

        // write current node, and set its parent
        long pointer = -1;
        try {
            pointer = oiom.writeObject(iNode);
        } catch (Exception e) {
            System.out.println("error writing index node to ObjectIOManager!");
            e.printStackTrace();
        }
        ;
        return pointer;
    }

    private class LoadTask {
        private List<? extends IndexObject> compressedData;
        private String                      description;
        private int                         parentNodeIndex;
        private int                         myIndex;

        private int                         size;
        private int                         numPivots;
        public List<IndexObject>            pivotHistoryList;
        private InternalNode                node;
        
        public String toString()
        {
            StringBuffer sb = new StringBuffer();
            sb.append("description=" + description + ", myIndex =" + myIndex + " parentindex=");
            sb.append(parentNodeIndex + ", size=" + size + ", numPivots=" + numPivots);
            if (node == null)
                sb.append(", node is null!");
            else
                sb.append(", node=" + node);
            return sb.toString();
        }

        LoadTask(List<? extends IndexObject> compressedData, String description, int parentIndex,
                int myIndex, List<IndexObject> pivotHistoryList) {
            this.compressedData = compressedData;
            this.description = description;
            this.parentNodeIndex = parentIndex;
            this.myIndex = myIndex;
            this.pivotHistoryList = pivotHistoryList;

            this.size = compressedData.size();
            this.numPivots = -1;
        }

        private void groupPivotsAtEnd(int[] pivots) {
            numPivots = pivots.length;
            if (numPivots < size) {
                int futurePivotIndex = size - 1;
                // put all the pivots at the end
                for (int i = 0; i < numPivots; i++) {
                    Collections.swap(compressedData, futurePivotIndex, pivots[i]);
                    // if we just swapped this pivot with another pivot, then we need to change that
                    // pivot's value
                    for (int j = i + 1; j < numPivots; j++) {
                        if (pivots[j] == futurePivotIndex) {
                            pivots[j] = pivots[i];
                        }
                    }
                    // pivots[i] = futurePivotIndex; don't need this since pivots[] is not used
                    // after this.
                    futurePivotIndex--;
                }
            }
        }

        // get the pivots (sans rowIDs); used as input for PivotSelectMethod.selectPivots()
        public IndexObject[] getPivots() {
            // TODO replace with context specific error
            if (numPivots == -1) {
                throw new Error("pivots have not yet been chosen!!!");
            }
            IndexObject[] pivots = new IndexObject[numPivots];
            int start = size - numPivots;
            for (int i = 0; i < numPivots; i++) {
                pivots[i] = compressedData.get(start);
                start++;
            }
            return pivots;
        }

        public IndexObject[] getDataPoints() {
            final int mySize;
            if (numPivots == -1)
                mySize = size;
            else
                mySize = size - numPivots;
            IndexObject[] dataPoints = new IndexObject[mySize];
            for (int i = 0; i < mySize; i++) {
                dataPoints[i] = compressedData.get(i);
            }
            return dataPoints;
        }
    }

    public Cursor search(Query q) {
        if (q instanceof RangeQuery)
            return new VPRangeCursor((RangeQuery) q, this);
        else if (q instanceof KNNQuery)
            return new VPKNNCursor((KNNQuery) q, this);
        else
            throw new UnsupportedOperationException("Unsupported query type " + q.getClass());
    }

	public ObjectIOManager getOIOM() {
		return oiom;
	}

	public long getRootAddress() {
		return root;
	}
}
