/**
 * mobios.index.VPLeafNode 2006.05.12
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.05.12: created, by Rui Mao, Willard
 */
package mobios.index;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import mobios.type.IndexObject;

/**
 * A VPLeafNode is a {@link LeafNode} in a {@link VPIndex}. 
 * 
 * @author Rui Mao, Willard
 */
public class VPLeafNode extends LeafNode {
    private static final long serialVersionUID = -2816878994630817839L;

    private double[][]        distance;
    private double[][]        pathDistance;

    public VPLeafNode() {
        super();
    }

    /**
     * Main constructor. 
     * 
     * @param pivots
     * 		an array of @{@link IndexObject}s comprising the pivots of this node.
     * @param children
     * 		an array of @{@link IndexObject}s comprising the children of this node.
     * @param size
     * 		the size of this leaf node.
     * @param distance
     *        distances from each distinct data point (row) to each pivot
     * @param pathDistance
     *        distances from each distinct data point (row) to each pivot(column) on the path from
     *        the root to this node.
     */
    public VPLeafNode(IndexObject[] pivots, IndexObject[] children, int size, double[][] distance,
            double[][] pathDistance) {
        super(pivots, children, size);

        if (distance == null)
            throw new IllegalArgumentException("distance array cannot be null!");
        this.distance = distance;

        if (pathDistance == null)
            throw new IllegalArgumentException("pathDistance array cannot be null!");

        this.pathDistance = pathDistance;
    }

    /**
     * @param childIndex
     * @return the distances from a child to all the pivots.
     */
    public double[] getDataPointPivotDistance(int childIndex) {
        return distance[childIndex];
    }

    /**
     * @param childIndex
     * @return the path distance list, consisting of the distances from a child to pivots on the
     *         path from the root to the node, can be of lenght 0, but can not be null.
     */
    public double[] getDataPointPathDistance(int childIndex) {
        return pathDistance[childIndex];
    }

    public void writeExternal(ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeInt(distance.length);
        for (int i = 0; i < distance.length; i++) {
            out.writeInt(distance[i].length);
            for (int j = 0; j < distance[i].length; j++) {
                out.writeDouble(distance[i][j]);
            }
        }
        out.writeInt(pathDistance.length);
        for (int i = 0; i < pathDistance.length; i++) {
            out.writeInt(pathDistance[i].length);
            for (int j = 0; j < pathDistance[i].length; j++) {
                out.writeDouble(pathDistance[i][j]);
            }
        }
    }

    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        distance = new double[in.readInt()][];
        for (int i = 0; i < distance.length; i++) {
            distance[i] = new double[in.readInt()];
            for (int j = 0; j < distance[i].length; j++) {
                distance[i][j] = in.readDouble();
            }
        }
        pathDistance = new double[in.readInt()][];
        for (int i = 0; i < pathDistance.length; i++) {
            pathDistance[i] = new double[in.readInt()];
            for (int j = 0; j < pathDistance[i].length; j++) {
                pathDistance[i][j] = in.readDouble();
            }
        }
    }
}
