/**
 * mobios.type.DoubleVector 2006.06.16 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.06.16: Modified from jdb 1.0, by Ru Mao
 */
package mobios.type;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Arrays;

import mobios.db.Table;
import mobios.db.TableManager;

/**
 * This class represents space vectors, where each element is a double.
 * 
 * @author Rui Mao, Willard
 * @version 2003.06.06
 */
public class DoubleVector extends IndexObject {
    
    /**TODO
     * 
     */
    private static final long serialVersionUID = 1275897899226331100L;

    /**
     * TODO
     */
    Table                     table;

    /** TODO
     * double array to store the data 
     */
    double[]                  data;

    /**
     * TODO
     */
    public DoubleVector() {
    }

    /**
     * Builds an instance from a space-separated {@link String} of doubles.
     * 
     * @param table
     * @param rowID
     * @param dataString
     */
    public DoubleVector(Table table, int rowID, String dataString) {
        super(rowID);
        this.table = table;
        String[] row = dataString.split("\\s+");
        data = new double[row.length];
        for (int i = 0; i < row.length; i++) {
            data[i] = Double.parseDouble(row[i]);
        }
    }

    /**
     * Builds an instance from a double array.
     * 
     * @param rowID
     * @param data
     *        the double array containning all the elements. cannot be null
     */
    public DoubleVector(Table table, int rowID, double[] data) {
        super(rowID);
        if (data == null)
            throw new IllegalArgumentException("null data constructing DoubleVector");
        this.table = table;
        this.data = (double[]) data.clone();
    }

    /**
     * @return the double array
     */
    public double[] getData() {
        return data;
    }

    /**
     * @return the dimension ( length) of the vector
     */
    public int size() {
        return data.length;
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.IndexObject#expand()
     */
    public IndexObject[] expand() {
        IndexObject[] dbO = new IndexObject[rowIDLength];
        for (int i = 0; i < rowIDLength; i++) {
            dbO[i] = new DoubleVector(table, rowIDStart + i, data);
        }
        return dbO;
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.IndexObject#compareTo(mobios.type.IndexObject)
     */
    public int compareTo(IndexObject oThat) {
        if (!(oThat instanceof DoubleVector))
            throw new ClassCastException("not compatible");
        return compareTo((DoubleVector) oThat);
    }

    /**
     * @param oThat
     * @return
     */
    public int compareTo(DoubleVector oThat) {
        DoubleVector that = (DoubleVector) oThat;
        if (this == that)
            return 0;

        if (this.size() < that.size())
            return -1;
        else if (this.size() > that.size())
            return 1;
        else {
            for (int i = 0; i < this.size(); i++) {
                double double1 = data[i];
                double double2 = that.data[i];
                if (double1 < double2)
                    return -1;
                else if (double1 > double2)
                    return 1;
            }
            return 0;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object that) {
        if (!(that instanceof DoubleVector))
            return false;
        return Arrays.equals(this.data, ((DoubleVector) that).data);
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#hashCode()
     */
    // taken from Joshua Bloch's Effective Java
    public int hashCode() {
        int result = 17;
        for (int i = 0; i < data.length; i++) {
            long _long = Double.doubleToLongBits(data[i]);
            result = 37 * result + (int) (_long ^ (_long >>> 32));
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("data: ");
        for (int i = 0; i < data.length; i++)
            sb.append(data[i]).append(", ");
        sb.append("rowID: "+table.getOriginalRowID(rowIDStart));
        return sb.toString();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
     */
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        super.readExternal(in);
        data = new double[in.readInt()];
        for (int i = 0; i < data.length; i++) {
            data[i] = in.readDouble();
        }
        table = TableManager.getTableManager().getTable(in.readInt());
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
     */
    public void writeExternal(ObjectOutput out) throws IOException {
        super.writeExternal(out);
        out.writeInt(data.length);
        for (int i = 0; i < data.length; i++) {
            out.writeDouble(data[i]);
        }
        out.writeInt(table.getTableLocation());
    }
}
