/**
 * mobios.type.Spectra 2003.08.13 
 * 
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log: 2003.08.13: Modified from the original mobios package, by Rui Mao 
 * 2003.11.06: Modified toString method for performance, by Willard 
 * 2004.05.13: Modified for performance, Willard 
 * 2004.05.12: Modified to use arrays instead of ArrayList, Willard
 * 2004.08.15: Modified constructor to fix initialization bug with String input data, by Smriti Ramakrishnan 
 * 2004.09.01: Added sortAsc() method, Smriti Ramakrishnan 
 * 2004.09.11: Added call to sortAsc() method in all constructors - all objects are are sorted mandatorily, by Smriti Ramakrishnan
 */
package mobios.type;

import java.util.Arrays;
import java.util.logging.Logger;

import mobios.db.Table;
import mobios.util.Debug;

/**
 * Represents a spectra. The only difference between a {@link Spectra} and a {@link DoubleVector} is
 * that the values in a <code>Spectra</code> are sorted from smallest to largest.
 * 
 * @author Smriti Ramakrishnan, Rui Mao, Willard
 */
public class Spectra extends DoubleVector {
    private static final long serialVersionUID = -8300375927493085758L;

    /**
     * Default Constructor.
     */
    public Spectra() {
    }

    /**
     * Constructor.
     * 
     * @param table
     *        the table this spectra comes from
     * @param rowID
     *        the rowID of this spectra in its table
     * @param spectra
     *        a space-separated {@link String} of doubles representing a spectra.
     */
    public Spectra(Table table, int rowID, String spectra) {
        super(table, rowID, spectra);
        // make sure query is sorted
        sortAsc();
    }

    /**
     * @param table
     *        the table this spectra comes from
     * @param rowID
     *        the rowID of this spectra in its table
     * @param spectra
     *        an array of doubles representing a spectra
     */
    public Spectra(Table table, int rowID, double[] spectra) {
        super(table, rowID, spectra);
        // make sure query is sorted
        sortAsc();
    }

    /**
     * Sorts the Spectra from smallest to largest values.
     */
    public void sortAsc() {
        try {
            Arrays.sort(data);
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (Debug.debug) {
            Logger.getLogger("mobios.index").finer("Sorted Spectra = " + toString());
        }
    }

    /**
     * @return the smallest value in the {@link Spectra}.
     */
    public double getMin() {
        return data[0];
    }

    /**
     * @return the largest value in the {@link Spectra}.
     */
    public double getMax() {
        return data[data.length - 1];
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.DoubleVector#expand()
     */
    public IndexObject[] expand() {
        IndexObject[] dbO = new IndexObject[rowIDLength];
        for (int i = 0; i < rowIDLength; i++) {
            dbO[i] = new Spectra(table, rowIDStart + i, data);
        }
        return dbO;
    }
}
