/**
 * mobios.type.DNA 2006.05.24
 *
 * The MoBIoS Library
 * Copyright (c) 2006 The MoBIoS Group and The University of Texas at Austin.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Change Log:
 * 2006.05.24: Modified from original jdb package, by Willard
 */
package mobios.type;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import mobios.dist.SymmetricSubstitutionWeightMatrix;
import mobios.dist.WeightMatrix;

/**
 * A <code>DNA</code> is a compact representation of a DNA sequence.
 * 
 * @author Rui Mao, Willard
 * @version 2003.06.06
 */
public class DNA extends Sequence {
    /**
     * 
     */
    private static final long serialVersionUID = 1351256505737923337L;

    /**
     * @author Willard
     */
    public static enum DNASymbol implements Symbol {
        A("Adenine", (byte) 0), C("Cytosine", (byte) 1), G("Guanine", (byte) 2), T("Thymine",
                (byte) 3), R("Purine", (byte) 4), Y("Pyrimidine", (byte) 5), M("C or A", (byte) 6), K(
                "T, U,or G", (byte) 7), W("T, U or A", (byte) 8), S("C or G", (byte) 9), B("not A",
                (byte) 10), D("not C", (byte) 11), H("not G", (byte) 12), V("not T, U", (byte) 13), N(
                "Any base", (byte) 14);

        /*
         * A description of the Symbol
         */
        private String description;
        
        /**
         * Every symbol has a non-unique byteValue. Some symbols may have the same byteValue
         * if there is no difference between them in the WeightMatrix.
         */
        public byte    byteValue;

        /**
         * 
         * 
         * @param description
         * @param byteValue
         */
        private DNASymbol(String description, byte byteValue) {
            this.description = description;
            this.byteValue = byteValue;
        }

        /**
         * 
         * @return the String description of the Symbol
         */
        public String description() {
            return description;
        }

        /* (non-Javadoc)
         * @see mobios.type.Symbol#byteValue()
         */
        public byte byteValue() {
            return byteValue;
        }

        /* (non-Javadoc)
         * @see mobios.type.Symbol#stringValue()
         */
        public String stringValue() {
            return toString();
        }

        /* (non-Javadoc)
         * @see mobios.type.Symbol#getSymbol(java.lang.String)
         */
        public Symbol getSymbol(String s) {
            return valueOf(s);
        }

        /* (non-Javadoc)
         * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
         */
        public void writeExternal(ObjectOutput out) throws IOException {
            out.writeInt(description.length());
            out.writeChars(description);
            out.writeByte(byteValue);
        }

        /* (non-Javadoc)
         * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
         */
        public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
            char[] charDescription = new char[in.readInt()];
            for (int i = 0; i < charDescription.length; i++) {
                charDescription[i] = in.readChar();
            }
            description = String.copyValueOf(charDescription);
            byteValue = in.readByte();

        }

        /**
         *
         * @return the number of Distinct byteValues for this set of Symbols.
         */
        public static int distinctSize() {
            return 15;
        }
    }

    /**
     * The alphabet of DNASymbols. Includes all of the symbols in DNASymbol.
     */
    public static final Alphabet ALPHABET        = new Alphabet(DNASymbol.values(), DNASymbol
                                                         .distinctSize());

    /**
     * A simple alphabet of DNASymbols. Only includes A,C, T, and G.
     */
    public static final Alphabet SIMPLE_ALPHABET = new Alphabet(new DNASymbol[] { DNASymbol.A,
            DNASymbol.C, DNASymbol.T, DNASymbol.G }, 4);

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.Sequence#getAlphabet()
     */
    public Alphabet getAlphabet() {
        return ALPHABET;
    }

    /**
     * Construct a DNA.
     * 
     * @param sequenceID
     *        the sequence ID of the DNA sequence. can be the accession number, or a user assigned
     *        ID.
     * @param sequence
     *        a string representing a DNA sequence.
     */
    public DNA(String sequenceID, String sequence) {
        super(sequenceID, sequence);
        for (int i = 0; i < data.length; i++) {
            data[i] = (DNASymbol.valueOf(sequence.substring(i, i + 1))).byteValue();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.Sequence#get(int)
     */
    public Symbol get(int index) {
        return ALPHABET.get(data[index]);
    }

    /*
     * (non-Javadoc)
     * 
     * @see mobios.type.Sequence#toString()
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer(data.length);
        for (int i = 0; i < data.length; i++)
            buffer.append(ALPHABET.get(data[i]));
        return buffer.toString();
    }

    /**
     * The SimpleDNAEditDistanceMatrix looks like: {0,1,1,1}, {1,0,1,1}, {1,1,0,1}, {1,1,1,0}
     */
    public final static WeightMatrix SimpleDNAEditDistanceMatrix = new SymmetricSubstitutionWeightMatrix(
                                                                         DNA.SIMPLE_ALPHABET,
                                                                         new double[][] {
            { 0, 1, 1, 1 }, 
            { 1, 0, 1, 1 }, 
            { 1, 1, 0, 1 }, 
            { 1, 1, 1, 0 } });

    /**
     * A simpleWeightedDNAMatrix.
     * 
     * { 0, 2 / 3, 2 / 3, 2 / 3 }, 
     * { 2 / 3, 0, 2 / 3, 2 / 3 }, 
     * { 2 / 3, 2 / 3, 0, 1 },
     * { 2 / 3, 2 / 3, 1, 0 } }
     */
    public final static WeightMatrix SimpleWeightedDNAMatrix     = new SymmetricSubstitutionWeightMatrix(
                                                                         DNA.SIMPLE_ALPHABET,
                                                                         new double[][] {
            { 0, 2 / 3, 2 / 3, 2 / 3 }, 
            { 2 / 3, 0, 2 / 3, 2 / 3 }, 
            { 2 / 3, 2 / 3, 0, 1 },
            { 2 / 3, 2 / 3, 1, 0 } });

    /**
     * A complete EditDistanceWeightMatrix:
     * 
     * { 0, 1, 1, 1, 0.5, 1, 0.5, 1, 0.5, 1, 1, 0.5, 0.5, 0.5, 0.5 },// A Adenine
     * { 1, 0, 1, 1, 1, 0.5, 0.5, 1, 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5 },// C Cytosine
     * { 1, 1, 0, 1, 0.5, 1, 1, 0.5, 1, 0.5, 0.5, 0.5, 1, 0.5, 0.5 },// G Guanine
     * { 1, 1, 1, 0, 1, 0.5, 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 1, 0.5 },// T Thymine
     * { 0.5, 1, 0.5, 1, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 }, // R Purine (A or G)
     * { 1, 0.5, 1, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// Y Pyrimidine (C, T, or U)
     * { 0.5, 0.5, 1, 1, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// M C or A
     * { 1, 1, 0.5, 0.5, 0.5, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// K T, U, or G
     * { 0.5, 1, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5 },// W T, U, or A
     * { 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5 },// S C or G
     * { 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5, 0.5, 0.5 },// B C, T, U, or G (not A)
     * { 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5, 0.5 },// D A, T, U, or G (not C)
     * { 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5 },// H A, T, U, or C (not G)
     * { 0.5, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5 },// V A, C, or G (not T, not U)
     * { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0 }// N Anybase (A,C,G,T,or U)
     */
    public final static WeightMatrix EditDistanceWeightMatrix    = new SymmetricSubstitutionWeightMatrix(
                                                                         DNA.ALPHABET,
                                                                         new double[][] {
            { 0, 1, 1, 1, 0.5, 1, 0.5, 1, 0.5, 1, 1, 0.5, 0.5, 0.5, 0.5 },// A Adenine
            { 1, 0, 1, 1, 1, 0.5, 0.5, 1, 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5 },// C Cytosine
            { 1, 1, 0, 1, 0.5, 1, 1, 0.5, 1, 0.5, 0.5, 0.5, 1, 0.5, 0.5 },// G Guanine
            { 1, 1, 1, 0, 1, 0.5, 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 1, 0.5 },// T Thymine
            { 0.5, 1, 0.5, 1, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 }, // R Purine (A or G)
            { 1, 0.5, 1, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// Y Pyrimidine (C, T, or U)
            { 0.5, 0.5, 1, 1, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// M C or A
            { 1, 1, 0.5, 0.5, 0.5, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 },// K T, U, or G
            { 0.5, 1, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 1, 0.5, 0.5, 0.5, 0.5, 0.5 },// W T, U, or A
            { 1, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 1, 0, 0.5, 0.5, 0.5, 0.5, 0.5 },// S C or G
            { 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5, 0.5, 0.5 },// B C, T, U, or G (not A)
            { 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5, 0.5 },// D A, T, U, or G (not C)
            { 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5, 0.5 },// H A, T, U, or C (not G)
            { 0.5, 0.5, 0.5, 1, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0, 0.5 },// V A, C, or G (not T, not U)
            { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0 }// N Anybase (A,C,G,T,or U)
                                                                         });

}
